#ifndef __CERT_HH__
#define __CERT_HH__

// copyright (C) 2002, 2003 graydon hoare <graydon@pobox.com>
// all rights reserved.
// licensed to the public under the terms of the GNU GPL (>= 2)
// see the file COPYING for details

#include "vocab.hh"
#include <vector>
#include <time.h>

// certs associate an opaque name/value pair with a particular identifier in
// the system (eg. a manifest or file id) and are accompanied by an RSA
// public-key signature attesting to the association. users can write as
// much extra meta-data as they like about files or manifests, using certs,
// without needing anyone's special permission.

struct app_state;

struct cert
{
  cert();
  cert(hexenc<id> const & ident,
      cert_name const & name,
      base64< gzip<cert_value> > const & value,
      rsa_keypair_id const & key);
  cert(hexenc<id> const & ident,
      cert_name const & name,
      base64< gzip<cert_value> > const & value,
      rsa_keypair_id const & key,
      hexenc<rsa_sha1_signature> const & sig);
  hexenc<id> ident;
  cert_name name;
  base64< gzip<cert_value> > value;
  rsa_keypair_id key;
  hexenc<rsa_sha1_signature> sig;
  bool operator<(cert const & other) const;
  bool operator==(cert const & other) const;
};

void cert_signable_text(cert const & t,
		       string & out);

bool check_cert(app_state & app, cert const & t);
void calculate_cert(app_state & app, cert & t);

// special certs -- system won't work without them

extern string const ancestor_cert_name;
extern string const branch_cert_name;

void cert_manifest_in_branch(manifest_id const & manifest, 
			     cert_value const & branchname,
			     app_state & app);

void get_branch_heads(cert_value const & branchname,
		      app_state & app,
		      vector<manifest_id> & heads);

void cert_file_ancestor(file_id const & parent, 
		       file_id const & child,
		       app_state & app);

void cert_manifest_ancestor(manifest_id const & parent, 
			   manifest_id const & child,
			   app_state & app);

bool find_common_ancestor(manifest_id const & left,
			  manifest_id const & right,
			  manifest_id & anc,
			  app_state & app);


// we also define some common cert types, to help establish useful
// conventions. you should use these unless you have a compelling
// reason not to.

bool guess_default_key(rsa_keypair_id & key, app_state & app);

extern string const date_cert_name;
extern string const author_cert_name;
extern string const petname_cert_name;
extern string const release_cert_name;
extern string const changelog_cert_name;
extern string const comment_cert_name;
extern string const approval_cert_name;
extern string const testresult_cert_name;

void cert_manifest_date_now(manifest_id const & m, 
			    app_state & app);

void cert_manifest_date_time(manifest_id const & m, 
			     time_t time,
			     app_state & app);

void cert_manifest_author(manifest_id const & m, 
			  string const & author,
			  app_state & app);

void cert_manifest_author_default(manifest_id const & m, 
				  app_state & app);

void cert_manifest_tag(manifest_id const & m, 
		       string const & tagname,
		       app_state & app);

void cert_manifest_changelog(manifest_id const & m, 
			     string const & changelog,
			     app_state & app);

void cert_file_comment(file_id const & m, 
		       string const & comment,
		       app_state & app);

void cert_manifest_comment(manifest_id const & m, 
			   string const & comment,
			   app_state & app);

void cert_file_approval(file_id const & m, 
			bool const approval,
			app_state & app);

void cert_manifest_approval(manifest_id const & m, 
			    bool const approval,
			    app_state & app);

void cert_manifest_testresult(manifest_id const & m, 
			      string const & suitename,
			      istream const & results,
			      app_state & app);



#endif // __CERT_HH__
