// copyright (C) 2002, 2003 graydon hoare <graydon@pobox.com>
// all rights reserved.
// licensed to the public under the terms of the GNU GPL (>= 2)
// see the file COPYING for details

#include <map>
#include <cstdio>
#include <set>
#include <vector>
#include <algorithm>
#include <iterator>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/lexical_cast.hpp>

#include "commands.hh"

#include "app_state.hh"
#include "diff_patch.hh"
#include "file_io.hh"
#include "keys.hh"
#include "manifest.hh"
#include "nntp_tasks.hh"
#include "packet.hh"
#include "patch_set.hh"
#include "rcs_import.hh"
#include "sanity.hh"
#include "cert.hh"
#include "transforms.hh"
#include "update.hh"
#include "vocab.hh"
#include "work.hh"

//
// this file defines the task-oriented "top level" commands which can be
// issued as part of a monotone command line. the command line can only
// have one such command on it, followed by a vector of strings which are its
// arguments. all --options will be processed by the main program *before*
// calling a command
//
// we might expose this blunt command interface to scripting someday. but
// not today.

namespace commands 
{
  struct command;
  bool operator<(command const & self, command const & other);
};

namespace std
{
  template <>
  struct std::greater<commands::command *>
  {
    bool operator()(commands::command const * a, commands::command const * b)
    {
      return *a < *b;
    }
  };
};

namespace commands 
{
using namespace std;

struct command; 

static map<string,command *> cmds;

struct command 
{
  string name;
  string group;
  string params;
  string desc;
  command(string const & n,
	  string const & g,
	  string const & p,
	  string const & d) : name(n), group(g), params(p), desc(d) 
  { cmds[n] = this; }
  virtual ~command() {}
  virtual void exec(app_state & app, vector<string> const & args) = 0;
};

bool operator<(command const & self, command const & other)
{
  return ((self.group < other.group)
	  || ((self.group == other.group) && (self.name < other.name)));
}


void explain_usage(ostream & out)
{
  vector<command *> sorted;
  out << "commands:" << endl;
  for (map<string,command *>::const_iterator i = cmds.begin();
       i != cmds.end(); ++i)
    {
      sorted.push_back(i->second);
    }
  
  sort(sorted.begin(), sorted.end(), std::greater<command *>());

  string curr_group;
  for (size_t i = 0; i < sorted.size(); ++i)
    {
      if (sorted[i]->group != curr_group)
	{
	  out << "[" << sorted[i]->group << " commands]" << endl << endl;
	  curr_group = sorted[i]->group;
	}
      out << "  -- " << sorted[i]->name << " " << sorted[i]->params << endl
	  << "     " << sorted[i]->desc << endl << endl;
    }
}

void process(app_state & app, string const & cmd, vector<string> const & args)
{
  if (cmds.find(cmd) != cmds.end())
    {
      L("executing %s command\n", cmd.c_str());
      cmds[cmd]->exec(app, args);
    }
  else
    {
      throw usage();
    }
}

#define CMD(C, group, params, desc)               \
struct cmd_ ## C : public command                 \
{                                                 \
  cmd_ ## C() : command(#C, group, params, desc)  \
  {}                                              \
  virtual void exec(app_state & app,              \
                    vector<string> const & args); \
};                                                \
static cmd_ ## C C ## _cmd;                       \
void cmd_ ## C::exec(app_state & app,             \
                     vector<string> const & args) \


static void ensure_bookdir()
{
  mkdir_p(local_path(book_keeping_dir));
}

static void get_manifest_path(local_path & m_path)
{
  m_path = (fs::path(book_keeping_dir) / fs::path(manifest_file_name)).string();
  L("manifest path is %s\n", m_path().c_str());
}

static void get_work_path(local_path & w_path)
{
  w_path = (fs::path(book_keeping_dir) / fs::path(work_file_name)).string();
  L("work path is %s\n", w_path().c_str());
}

static void get_manifest_map(manifest_map & m)
{
  ensure_bookdir();
  local_path m_path;
  base64< gzip<data> > m_data;
  get_manifest_path(m_path);
  if (file_exists(m_path))
    {
      L("loading manifest file %s\n", m_path().c_str());      
      read_data(m_path, m_data);
      read_manifest_map(manifest_data(m_data), m);
      L("read %d manifest entries\n", m.size());
    }
  else
    {
      L("no manifest file %s\n", m_path().c_str());
    }
}

static void put_manifest_map(manifest_map const & m)
{
  ensure_bookdir();
  local_path m_path;
  manifest_data m_data;
  get_manifest_path(m_path);
  L("writing manifest file %s\n", m_path().c_str());
  write_manifest_map(m, m_data);
  write_data(m_path, m_data.inner());
  L("wrote %d manifest entries\n", m.size());
}

static void get_work_set(work_set & w)
{
  ensure_bookdir();
  local_path w_path;
  get_work_path(w_path);
  if (file_exists(w_path))
    {
      L("checking for un-committed work file %s\n", 
	w_path().c_str());
      data w_data;
      read_data(w_path, w_data);
      read_work_set(w_data, w);
      L("read %d dels, %d adds from %s\n", 
	w.dels.size(), w.adds.size(), w_path().c_str());
    }
  else
    {
      L("no un-committed work file %s\n", w_path().c_str());
    }
}

static void put_work_set(work_set & w)
{
  local_path w_path;
  get_work_path(w_path);

  if (w.adds.size() > 0
      || w.dels.size() > 0)
    {
      ensure_bookdir();
      data w_data;
      write_work_set(w_data, w);
      write_data(w_path, w_data);
    }
  else
    {
      delete_file(w_path);
    }
}

static void calculate_new_manifest_map(manifest_map const & m_old, 
				       manifest_map & m_new)
{
  path_set paths;
  work_set work;
  extract_path_set(m_old, paths);
  get_work_set(work);
  if (work.dels.size() > 0)
    L("removing %d dead files from manifest\n", 
      work.dels.size());
  if (work.adds.size() > 0)
    L("adding %d files to manifest\n", work.adds.size());
  apply_work_set(work, paths);
  build_manifest_map(paths, m_new);
}

static string get_stdin()
{
  size_t const bufsz = 0xffff;
  char buf[bufsz];
  string tmp;
  while(cin)
    {
      cin.read(buf, bufsz);
      tmp.append(buf, cin.gcount());
    }
  return tmp;
}

static void get_log_message(patch_set const & ps, 
			    app_state & app,
			    string & log_message)
{
  string commentary;
  string summary;
  stringstream ss;
  patch_set_to_text_summary(ps, ss);
  summary = ss.str();
  commentary += "----------------------------------------------------------------------\n";
  commentary += "Enter Log.  Lines beginning with `MT:' are removed automatically\n";
  commentary += "\n";
  commentary += "Summary of changes:\n";
  commentary += "\n";
  commentary += summary;
  commentary += "----------------------------------------------------------------------\n";
  if (!app.lua.hook_edit_comment(commentary, log_message))
    throw oops("edit of log message failed");
}


// actual commands follow

CMD(lscerts, "key and cert", "(file|manifest) <id>", 
    "list certs associated with manifest or file")
{
  if (args.size() != 2)
    throw usage();

  vector<cert> certs;

  transaction_guard guard(app.db);

  if (args[0] == "manifest")
    {
      manifest_id ident(args[1]);
      vector< manifest<cert> > ts;
      app.db.get_manifest_certs(ident, ts);
      for (size_t i = 0; i < ts.size(); ++i)
	certs.push_back(ts[i].inner());
    }
  else if (args[0] == "file")
    {
      file_id ident(args[1]);
      vector< file<cert> > ts;
      app.db.get_file_certs(ident, ts);
      for (size_t i = 0; i < ts.size(); ++i)
	certs.push_back(ts[i].inner());
    }
  else
    throw usage();
	
  for (size_t i = 0; i < certs.size(); ++i)
    {
      bool ok = check_cert(app, certs[i]);
      gzip<cert_value> decoded;
      cert_value tv;      
      decode_base64(certs[i].value, decoded);
      decode_gzip(decoded, tv);      
      cout << (ok ? "ok " : "bad ")
	   << "[" << certs[i].key() << "] " 
	   << certs[i].name() << " = " 
	   << tv() << endl;
    }

  guard.commit();
}

CMD(lskeys, "key and cert", "[partial-id]", "list keys")
{
  vector<rsa_keypair_id> pubkeys;
  vector<rsa_keypair_id> privkeys;

  transaction_guard guard(app.db);

  if (args.size() == 0)
    app.db.get_key_ids("", pubkeys, privkeys);
  else if (args.size() == 1)
    app.db.get_key_ids(args[0], pubkeys, privkeys);
  else
    throw usage();
  
  if (pubkeys.size() > 0)
    {
      cout << endl << "[public keys]" << endl;
      for (size_t i = 0; i < pubkeys.size(); ++i)
	cout << pubkeys[i]() << endl;
      cout << endl;
    }

  if (privkeys.size() > 0)
    {
      cout << endl << "[private keys]" << endl;
      for (size_t i = 0; i < privkeys.size(); ++i)
	cout << privkeys[i]() << endl;
      cout << endl;
    }

  guard.commit();
}

CMD(genkey, "key and cert", "<keyid>", "generate an RSA key-pair")
{
  if (args.size() != 1)
    throw usage();
  
  transaction_guard guard(app.db);
  rsa_keypair_id ident(args[0]);

  if (app.db.key_exists(ident))
    throw oops("key '" + ident() + "' already exists in database");
  
  base64<rsa_pub_key> pub;
  base64< arc4<rsa_priv_key> > priv;
  L("generating key-pair '%s'\n", ident().c_str());
  generate_key_pair(app.lua, ident, pub, priv);
  L("storing key-pair '%s' in database\n", ident().c_str());
  app.db.put_key_pair(ident, pub, priv);

  guard.commit();
}

CMD(cert, "key and cert", "(file|manifest) <id> <certname> [certval]", 
                        "create a cert for a file or manifest")
{
  if ((args.size() != 4) && (args.size() != 3))
    throw usage();

  transaction_guard guard(app.db);

  hexenc<id> ident(args[1]);
  cert_name name(args[2]);

  rsa_keypair_id key;
  if (app.signing_key() != "")
    key = app.signing_key;
  else
    if (!guess_default_key(key, app))
      throw oops("no unique private key found, and no key specified");
  
  cert_value val;
  if (args.size() == 4)
    val = cert_value(args[3]);
  else
    val = cert_value(get_stdin());

  gzip<cert_value> val_compressed;
  base64< gzip<cert_value> > val_encoded;
  encode_gzip(val, val_compressed);
  encode_base64(val_compressed, val_encoded);

  cert t(ident, name, val_encoded, key);
  
  // nb: we want to throw usage on mis-use *before* asking for a
  // passphrase.

  if (args[0] == "file")
    {
      calculate_cert(app, t);
      app.db.put_file_cert(file<cert>(t));
    }
  else if (args[0] == "manifest")
    {
      calculate_cert(app, t);
      app.db.put_manifest_cert(manifest<cert>(t));
    }
  else
    throw usage();

  guard.commit();
}


CMD(tag, "certificate", "<id> <tagname>", 
    "put a symbolic tag cert on a manifest version")
{
  if (args.size() != 2)
    throw usage();
  manifest_id m(args[0]);
  cert_manifest_tag(m, args[1], app);
}

CMD(approve, "certificate", "(file|manifest) <id>", 
    "approve of a manifest or file version")
{
  if (args.size() != 2)
    throw usage();
  if (args[0] == "manifest")
    {
      manifest_id m(args[1]);
      cert_manifest_approval(m, true, app);
    }
  else if (args[0] == "file")
    {
      file_id f(args[1]);
      cert_file_approval(f, true, app);
    }
  else
    throw usage();
}

CMD(disapprove, "certificate", "(file|manifest) <id>", 
    "disapprove of a manifest or file version")
{
  if (args.size() != 2)
    throw usage();
  if (args[0] == "manifest")
    {
      manifest_id m(args[1]);
      cert_manifest_approval(m, false, app);
    }
  else if (args[0] == "file")
    {
      file_id f(args[1]);
      cert_file_approval(f, false, app);
    }
  else
    throw usage();
}


CMD(comment, "certificate", "(file|manifest) <id> [comment]", 
    "comment on a file or manifest version")
{
  if (args.size() != 2 && args.size() != 3)
    throw usage();

  string comment;
  if (args.size() == 3)
    comment = args[2];
  else
    if (!app.lua.hook_edit_comment("", comment))
      throw oops("edit comment failed");

  if (comment.find_first_not_of(" \r\t\n") == string::npos)
    throw oops("empty comment");

  if (args[0] == "file")
    cert_file_comment(file_id(args[1]), comment, app); 
  else if (args[0] == "manifest")
    cert_manifest_comment(manifest_id(args[1]), comment, app);
  else
    throw usage();
}



CMD(add, "working copy", "<pathname> [...]", "add files to working copy")
{
  if (args.size() < 1)
    throw usage();

  transaction_guard guard(app.db);

  manifest_map man;
  work_set work;  
  get_manifest_map(man);
  get_work_set(work);
  bool rewrite_work = false;

  for (vector<string>::const_iterator i = args.begin(); i != args.end(); ++i)
    {
      if (man.find(*i) != man.end())
	{
	  L("skipping %s, already present in manifest\n", i->c_str());
	  continue;
	}

      if (work.adds.find(*i) != work.adds.end())
	{
	  L("skipping %s, already present in working copy add set\n", i->c_str());
	  continue;
	}

      if (work.dels.find(*i) != work.dels.end())
	{
	  L("removing %s from working copy delete set\n", i->c_str());
	  work.dels.erase(*i);
	  rewrite_work = true;
	}
      else
	{
	  L("adding %s to working copy add set\n", i->c_str());
	  work.adds.insert(*i);
	  rewrite_work = true;
	}
    }
  guard.commit();

  // small race here
  if (rewrite_work)
    put_work_set(work);
}

CMD(drop, "working copy", "<pathname> [...]", "drop files from working copy")
{
  if (args.size() < 1)
    throw usage();

  manifest_map man;
  work_set work;
  get_manifest_map(man);
  get_work_set(work);
  bool rewrite_work = false;

  transaction_guard guard(app.db);

  for (vector<string>::const_iterator i = args.begin(); i != args.end(); ++i)
    {
      if (man.find(*i) == man.end())
	{
	  L("skipping %s, does not exist in manifest\n", i->c_str());
	  continue;
	}

      if (work.dels.find(*i) != work.dels.end())
	{
	  L("skipping %s, already present in working copy delete set\n", i->c_str());
	  continue;
	}

      if (work.adds.find(*i) != work.adds.end())
	{
	  L("removing %s from working copy add set\n", i->c_str());
	  work.adds.erase(*i);
	  rewrite_work = true;
	}
      else
	{
	  L("adding %s to working copy delete set\n", i->c_str());
	  work.dels.insert(*i);
	  rewrite_work = true;
	}
    }

  guard.commit();

  // small race here
  if (rewrite_work)
    put_work_set(work);
}

CMD(commit, "working copy", "[log message]", "commit working copy to database")
{
  string log_message("");
  manifest_map m_old, m_new;
  patch_set ps;

  get_manifest_map(m_old);
  calculate_new_manifest_map(m_old, m_new);
  manifest_id old_id, new_id;
  calculate_manifest_map_ident(m_old, old_id);
  calculate_manifest_map_ident(m_new, new_id);

  if (args.size() != 0 && args.size() != 1)
    throw usage();
  
  cert_value branchname;
  if (app.branch_name != "")
    {
      branchname = app.branch_name;
    }
  else
    {
      vector< manifest<cert> > certs;
      cert_name branch(branch_cert_name);
      app.db.get_manifest_certs(old_id, branch, certs);
      if (certs.size() == 0)
	throw oops("no branch certs found for old manifest " 
		   + old_id.inner()() + ", please provide a branch name");
      else if (certs.size() > 1)
	throw oops("multiple branch certs found for old manifest " 
		   + old_id.inner()() + ", please provide a branch name");
      else
	unpack(certs[0].inner().value, branchname);
    }
    
  L("committing %s to branch %s\n", 
    new_id.inner()().c_str(), branchname().c_str());
  app.branch_name = branchname();

  manifests_to_patch_set(m_old, m_new, app, ps);

  // get log message
  if (args.size() == 1)
    log_message = args[0];
  else
    get_log_message(ps, app, log_message);

  if (log_message.find_first_not_of(" \r\t\n") == string::npos)
    throw oops("empty log message");

  {
    transaction_guard guard(app.db);

    // process manifest delta or new manifest
    if (app.db.manifest_version_exists(ps.m_new))
      {
	L("skipping manifest %s, already in database\n", ps.m_new.inner()().c_str());
      }
    else
      {
	if (app.db.manifest_version_exists(ps.m_old))
	  {
	    L("inserting manifest delta %s -> %s\n", 
	      ps.m_old.inner()().c_str(), ps.m_new.inner()().c_str());
	    manifest_data m_old_data, m_new_data;
	    app.db.get_manifest_version(ps.m_old, m_old_data);
	    write_manifest_map(m_new, m_new_data);
	    base64< gzip<delta> > del;
	    diff(m_old_data.inner(), m_new_data.inner(), del);
	    app.db.put_manifest_version(ps.m_old, ps.m_new, manifest_delta(del));
	  }
	else
	  {
	    L("inserting full manifest %s\n", 
	      ps.m_new.inner()().c_str());
	    manifest_data m_new_data;
	    write_manifest_map(m_new, m_new_data);
	    app.db.put_manifest(ps.m_new, m_new_data);
	  }
      }

    // process file deltas
    for (set<patch_delta>::const_iterator i = ps.f_deltas.begin();
	 i != ps.f_deltas.end(); ++i)
      {
	if (app.db.file_version_exists(i->id_new))
	  {
	    L("skipping file delta %s, already in database\n", i->id_new.inner()().c_str());
	  }
	else
	  {
	    if (app.db.file_version_exists(i->id_old))
	      {
		L("inserting delta %s -> %s\n", 
		  i->id_old.inner()().c_str(), i->id_new.inner()().c_str());
		file_data old_data;
		base64< gzip<data> > new_data;
		app.db.get_file_version(i->id_old, old_data);
		read_data(i->path, new_data);
		base64< gzip<delta> > del;
		diff(old_data.inner(), new_data, del);
		app.db.put_file_version(i->id_old, i->id_new, file_delta(del));
	      }
	    else
	      {
		L("inserting full version %s\n", i->id_old.inner()().c_str());
		base64< gzip<data> > new_data;
		read_data(i->path, new_data);
		// sanity check
		hexenc<id> tid;
		calculate_ident(new_data, tid);
		I(tid == i->id_new.inner());
		app.db.put_file(i->id_new, file_data(new_data));
	      }
	  }
      }
  
    // process file adds
    for (set<patch_addition>::const_iterator i = ps.f_adds.begin();
	 i != ps.f_adds.end(); ++i)
      {
	if (app.db.file_version_exists(i->ident))
	  {
	    L("skipping file %s %s, already in database\n", 
	      i->path().c_str(), i->ident.inner()().c_str());
	  }
	else
	  {
	    // it's a new file
	    L("inserting new file %s %s\n", 
	      i->path().c_str(), i->ident.inner()().c_str());
	    base64< gzip<data> > new_data;
	    read_data(i->path, new_data);
	    app.db.put_file(i->ident, new_data);
	  }
      }

    cert_manifest_ancestor(ps.m_old, ps.m_new, app);
    cert_manifest_in_branch(ps.m_new, branchname, app); 
    cert_manifest_date_now(ps.m_new, app);
    cert_manifest_author_default(ps.m_new, app);
    cert_manifest_changelog(ps.m_new, log_message, app);

    // commit done, now queue diff for sending

    if (app.db.manifest_version_exists(ps.m_old)
	&& app.db.manifest_version_exists(ps.m_new))
    {
      server serv;
      newsgroup grp;
      if (app.lua.hook_get_newsserver(branchname, serv) 
	  && app.lua.hook_get_newsgroup(branchname, grp))
	{
	  posting post(serv, grp, ps.m_old, ps.m_new);
	  app.db.queue_to_send(post);
	  cout << "queued to send to " 
	       << post.serv() << " : "
	       << post.group() << endl;
	}
      else
	{
	  L("no news server / news group found\n");
	}
    }
    guard.commit();
  }
  // small race condition here...
  local_path w_path;
  get_work_path(w_path);
  delete_file(w_path);
  put_manifest_map(m_new);
  cout << "committed " << ps.m_new.inner()() << endl;
}

CMD(update, "working copy", "[sort keys...]", "update working copy, relative to sorting keys")
{

  manifest_data m_chosen_data;
  manifest_map m_old, m_working, m_chosen, m_new;
  manifest_id m_old_id, m_chosen_id;

  transaction_guard guard(app.db);

  get_manifest_map(m_old);
  calculate_manifest_map_ident(m_old, m_old_id);
  calculate_new_manifest_map(m_old, m_working);
  
  pick_update_target(m_old_id, args, app, m_chosen_id);
  L("selected update target %s\n",
    m_chosen_id.inner()().c_str());
  app.db.get_manifest_version(m_chosen_id, m_chosen_data);
  read_manifest_map(m_chosen_data, m_chosen);

  update_merge_provider merger(app);
  if (!merge3(m_old, m_chosen, m_working, app, merger, m_new))
    throw oops("manifest merge failed, no update performed");

  L("calculating patchset for update\n");
  patch_set ps;
  manifests_to_patch_set(m_working, m_new, app, ps);

  L("applying %d deletions to files in tree\n", ps.f_dels.size());
  for (set<file_path>::const_iterator i = ps.f_dels.begin();
       i != ps.f_dels.end(); ++i)
    {
      L("deleting %s\n", (*i)().c_str());
      delete_file(*i);
    }

  L("applying %d moves to files in tree\n", ps.f_moves.size());
  for (set<patch_move>::const_iterator i = ps.f_moves.begin();
       i != ps.f_moves.end(); ++i)
    {
      L("moving %s -> %s\n", i->path_old().c_str(), i->path_new().c_str());
      move_file(i->path_old, i->path_new);
    }
  
  L("applying %d additions to tree\n", ps.f_adds.size());
  for (set<patch_addition>::const_iterator i = ps.f_adds.begin();
       i != ps.f_adds.end(); ++i)
    {
      L("adding %s as %s\n", i->ident.inner()().c_str(), i->path().c_str());
      file_data tmp;
      if (app.db.file_version_exists(i->ident))
	app.db.get_file_version(i->ident, tmp);
      else if (merger.temporary_store.find(i->ident) != merger.temporary_store.end())
	tmp = merger.temporary_store[i->ident];
      else
	I(false); // trip assert. this should be impossible.
      write_data(i->path, tmp.inner());
    }

  L("applying %d deltas to tree\n", ps.f_deltas.size());
  for (set<patch_delta>::const_iterator i = ps.f_deltas.begin();
       i != ps.f_deltas.end(); ++i)
    {
      L("altering file %s: %s -> %s\n", 
	i->path().c_str(),
	i->id_old.inner()().c_str(),
	i->id_new.inner()().c_str());

      // sanity check
      {
	base64< gzip<data> > dtmp;
	hexenc<id> dtmp_id;
	read_data(i->path, dtmp);
	calculate_ident(dtmp, dtmp_id);
	I(dtmp_id == i->id_old.inner());
      }

      // ok, replace with new version
      {
	file_data tmp;
	if (app.db.file_version_exists(i->id_new))
	  app.db.get_file_version(i->id_new, tmp);
	else if (merger.temporary_store.find(i->id_new) != merger.temporary_store.end())
	  tmp = merger.temporary_store[i->id_new];
	else
	  I(false); // trip assert. this should be impossible.
	write_data(i->path, tmp.inner());
      }
    }
  
  L("update successful\n");
  guard.commit();

  // small race condition here...
  // nb: we write out m_chosen, not m_new, because the manifest-on-disk
  // is the basis of the working copy, not the working copy itself.
  put_manifest_map(m_chosen);
  cout << "updated to base version " << m_chosen_id.inner()() << endl;
}


CMD(scan, "tree", "[pathnames...]", "scan files and generate a manifest for working copy")
{
  manifest_map m;
  if (args.size() > 0)
    {
      for (size_t i = 0; i < args.size(); ++i)
	{
	  manifest_map tmp;
	  L("building manifest for %s\n", args[i].c_str());
	  build_manifest_map(file_path(args[i]), app, tmp);
	  append_manifest_map(tmp, m);
	}
    }
  else
    {
      L("building manifest for cwd\n");
      build_manifest_map(app, m);
    }
  put_manifest_map(m);
}


CMD(cat, "tree", "(file|manifest) <id>", "write file or manifest from database to stdout")
{
  if (args.size() != 2)
    throw usage();

  transaction_guard guard(app.db);

  if (args[0] == "file")
    {
      file_data dat;
      file_id ident(args[1]);
      
      if (app.db.file_version_exists(ident))
	{
	  L("dumping file %s\n", ident.inner()().c_str());
	  app.db.get_file_version(ident, dat);
	  data unpacked;
	  unpack(dat.inner(), unpacked);
	  cout.write(unpacked().data(), unpacked().size());
	}
      else
	{
	  throw oops("no file version " + ident.inner()() + " found in database");
	}
    }
  else if (args[0] == "manifest")
    {
      manifest_data dat;
      manifest_id ident(args[1]);
      
      if (app.db.manifest_version_exists(ident))
	{
	  L("dumping manifest %s\n", ident.inner()().c_str());
	  app.db.get_manifest_version(ident, dat);
	  cout.write(dat.inner()().data(), dat.inner()().size());
	}
      else
	{
	  throw oops("no manifest version " + ident.inner()() + " found in database");
	}
    }
  else throw usage();

  guard.commit();
}


CMD(checkout, "tree", "<manifest-id>", "check out tree state from database")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);

  file_data data;
  manifest_id ident(args[0]);
  manifest_map m;

  if (app.db.manifest_version_exists(ident))
    {
      L("exporting manifest %s\n", ident.inner()().c_str());
      manifest_data m_data;
      app.db.get_manifest_version(ident, m_data);
      read_manifest_map(m_data, m);      
      put_manifest_map(m);

      for (manifest_map::const_iterator i = m.begin(); i != m.end(); ++i)
	{
	  vector<string> args;
	  path_id_pair pip(*i);

	  if (app.db.file_version_exists(pip.ident()))
	    {
	      file_data dat;
	      L("writing file %s to %s\n", 
		pip.ident().inner()().c_str(),
		pip.path()().c_str());
	      app.db.get_file_version(pip.ident(), dat);
	      write_data(pip.path(), dat.inner());
	    }
	  else
	    {
	      throw oops("no file version " 
			 + pip.ident().inner()() 
			 + " found in database");
	    }
	}
    }
  else
    {
      throw oops("no manifest version " + ident.inner()() + " found in database");
    }

  guard.commit();
}

CMD(heads, "tree", "", "show unmerged heads of branch")
{
  vector<manifest_id> heads;
  if (args.size() != 0)
    throw usage();

  if (app.branch_name == "")
    {
      cout << "please specify a branch, with --branch=<branchname>" << endl;
      return;
    }

  get_branch_heads(app.branch_name, app, heads);

  if (heads.size() == 0)
    cout << "branch '" << app.branch_name << "' is empty" << endl;
  else if (heads.size() == 1)
    cout << "branch '" << app.branch_name << "' is currently merged:" << endl;
  else
    cout << "branch '" << app.branch_name << "' is currently unmerged:" << endl;

  for (vector<manifest_id>::const_iterator i = heads.begin(); 
       i != heads.end(); ++i)
    {
      cout << i->inner()() << endl;
    }
}



CMD(merge, "tree", "", "merge unmerged heads of branch")
{

  vector<manifest_id> heads;

  if (args.size() != 0)
    throw usage();

  if (app.branch_name == "")
    {
      cout << "please specify a branch, with --branch=<branchname>" << endl;
      return;
    }

  get_branch_heads(app.branch_name, app, heads);

  if (heads.size() == 0)
    {
      cout << "branch " << args[0] << "is empty" << endl;
      return;
    }
  else if (heads.size() == 1)
    {
      cout << "branch " << args[0] << "is merged" << endl;
      return;
    }
  else
    {
      manifest_id left = heads[0];
      for (size_t i = 1; i < heads.size(); ++i)
	{
	  manifest_id right = heads[i];
	  manifest_id ancestor;
	  L("merge stage: %s <-> %s\n",
	    left.inner()().c_str(), right.inner()().c_str());
	  if(!find_common_ancestor(left, right, ancestor, app))
	    {
	      cout << "no common ancestor for" 
		   << left.inner()()
		   << " and "
		   << right.inner()()
		   << ", skipping" << endl;
	      continue;
	    }
	  L("common ancestor is %s\n", ancestor.inner()().c_str());
	  manifest_data left_data, right_data, ancestor_data, merged_data;
	  manifest_map left_map, right_map, ancestor_map, merged_map;
	  manifest_id merged;

	  app.db.get_manifest_version(left, left_data);
	  app.db.get_manifest_version(right, right_data);
	  app.db.get_manifest_version(ancestor, ancestor_data);

	  read_manifest_map(left_data, left_map);
	  read_manifest_map(right_data, right_map);
	  read_manifest_map(ancestor_data, ancestor_map);
	 
	  simple_merge_provider merger(app);
	  if (!merge3(ancestor_map, left_map, right_map, 
		      app, merger, merged_map))
	    {
	      throw oops("failed to merge manifests " 
			 + left.inner()() 
			 + " and " 
			 + right.inner()());
	    }
	  else
	    {
	      // merged 1 edge; now we commit this, update merge source and
	      // try next one
	      transaction_guard guard(app.db);
	      write_manifest_map(merged_map, merged_data);
	      calculate_manifest_map_ident(merged_map, merged);

	      base64< gzip<delta> > left_edge;
	      diff(left_data.inner(), merged_data.inner(), left_edge);
	      app.db.put_manifest_version(left, merged, left_edge);

	      // FIXME: we do *not* manufacture or store the second edge to
	      // the merged version, since doing so violates the
	      // assumptions of the db, and the 'right' version already
	      // exists in its entirety, anyways. this is a subtle issue
	      // though and I'm not sure I'm making the right
	      // decision. revisit. if you do not see that it is a subtle
	      // issue I suggest you are not thinking about it long enough.
	      //
	      // base64< gzip<delta> > right_edge;
	      // diff(right_data.inner(), merged_data.inner(), right_edge);
	      // app.db.put_manifest_version(right, merged, right_edge);


	      // we do of course record the ancestry relationship to both
	      // predecessors.

	      cert_manifest_ancestor(left, merged, app);
	      cert_manifest_ancestor(right, merged, app);
	      cert_manifest_in_branch(merged, app.branch_name, app);

	      cert_manifest_date_now(merged, app);
	      cert_manifest_author_default(merged, app);
	      cert_manifest_changelog(merged, 
				      "merge of " 
				      + left.inner()() 
				      + " and " 
				      + right.inner()(), app);
	      
	      guard.commit();
	      cout << "[source] " << left.inner()()   << endl
		   << "[source] " << right.inner()()  << endl
		   << "[merged] " << merged.inner()() << endl;
	      left = merged;
	    }
	}
    }
}


CMD(import, "tree", "[pathnames...]", "import file set and/or manifest into database")
{
  path_set ps;

  transaction_guard guard(app.db);

  if (args.size() < 1)
    {
      // we are importing everything in the working copy
      file_data m_data;
      manifest_id m_id;
      manifest_map m_old, m_new;
      
      get_manifest_map(m_old);
      calculate_new_manifest_map(m_old, m_new);
      
      calculate_manifest_map_ident(m_new, m_id);
      if (app.db.manifest_version_exists(m_id))
	{
	  L("skipping manifest %s, already present in db\n", m_id.inner()().c_str());
	}
      else
	{
	  L("importing manifest %s\n", m_id.inner()().c_str());
	  manifest_data m_data;
	  write_manifest_map(m_new, m_data);
	  app.db.put_manifest(m_id, m_data);
	}
      extract_path_set(m_new, ps);

      if (app.branch_name != "")
	  cert_manifest_in_branch(m_id, app.branch_name, app);

      cert_manifest_date_now(m_id, app);
      cert_manifest_author_default(m_id, app);
      cert_manifest_changelog(m_id, "import", app);

      cout << "imported manifest " << m_id.inner()() << endl;
    }
  else
    {
      // we are importing files explicitly mentionned on the command line
      copy(args.begin(), args.end(), inserter(ps, ps.begin()));
    }
    
  // now import the files
  for(path_set::const_iterator i = ps.begin(); i != ps.end(); ++i)
    {
      
      base64< gzip<data> > dat;
      hexenc<id> ident;
      read_data(*i, dat);
      calculate_ident(dat, ident);

      if (app.db.file_version_exists(file_id(ident)))
	{
	  L("skipping file %s (%s), already in database\n", (*i)().c_str(), ident().c_str());
	  continue;
	}
      app.db.put_file(file_id(ident), file_data(dat));
    }
  guard.commit();
}


CMD(diff, "informative", "", "show current diffs on stdout")
{
  manifest_map m_old, m_new;
  patch_set ps;

  transaction_guard guard(app.db);

  get_manifest_map(m_old);
  calculate_new_manifest_map(m_old, m_new);
  manifests_to_patch_set(m_old, m_new, app, ps);

  for (set<patch_delta>::const_iterator i = ps.f_deltas.begin();
       i != ps.f_deltas.end(); ++i)
    {
      file_data f_old;
      gzip<data> decoded_old;
      data decompressed_old, decompressed_new;
      vector<string> old_lines, new_lines;

      app.db.get_file_version(i->id_old, f_old);
      decode_base64(f_old.inner(), decoded_old);
      decode_gzip(decoded_old, decompressed_old);

      read_data(i->path, decompressed_new);

      split_into_lines(decompressed_old(), old_lines);
      split_into_lines(decompressed_new(), new_lines);

      unidiff(i->path(), i->path(), old_lines, new_lines, cout);
    }  
  guard.commit();
}

CMD(status, "informative", "", "show status of working copy")
{
  manifest_map m_old, m_new;
  patch_set ps;

  transaction_guard guard(app.db);
  get_manifest_map(m_old);
  calculate_new_manifest_map(m_old, m_new);
  manifests_to_patch_set(m_old, m_new, app, ps);
  patch_set_to_text_summary(ps, cout);
  guard.commit();
}


CMD(mdelta, "packet i/o", "<oldid> <newid>", "write manifest delta packet to stdout")
{
  if (args.size() != 2)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  manifest_id m_old_id, m_new_id; 
  manifest_data m_old_data, m_new_data;
  manifest_map m_old, m_new;
  patch_set ps;      
  m_old_id = hexenc<id>(args[0]); 
  m_new_id = hexenc<id>(args[1]);
  app.db.get_manifest_version(m_old_id, m_old_data);
  app.db.get_manifest_version(m_new_id, m_new_data);
  read_manifest_map(m_old_data, m_old);
  read_manifest_map(m_new_data, m_new);
  manifests_to_patch_set(m_old, m_new, app, ps);
  patch_set_to_packets(ps, app, pw);  
  guard.commit();
}

CMD(fdelta, "packet i/o", "<oldid> <newid>", "write file delta packet to stdout")
{
  if (args.size() != 2)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  file_id f_old_id, f_new_id;
  file_data f_old_data, f_new_data;
  f_old_id = hexenc<id>(args[0]);
  f_new_id = hexenc<id>(args[1]);     
  app.db.get_file_version(f_old_id, f_old_data);
  app.db.get_file_version(f_new_id, f_new_data);
  base64< gzip<delta> > del;
  diff(f_old_data.inner(), f_new_data.inner(), del);
  pw.consume_file_delta(f_old_id, f_new_id, file_delta(del));  
  guard.commit();
}

CMD(mdata, "packet i/o", "<id>", "write manifest data packet to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  manifest_id m_id;
  manifest_data m_data;
  m_id = hexenc<id>(args[0]);
  app.db.get_manifest_version(m_id, m_data);
  pw.consume_manifest_data(m_id, m_data);  
  guard.commit();
}


CMD(fdata, "packet i/o", "<id>", "write file data packet to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  file_id f_id;
  file_data f_data;
  f_id = hexenc<id>(args[0]);
  app.db.get_file_version(f_id, f_data);
  pw.consume_file_data(f_id, f_data);  
  guard.commit();
}

CMD(mcerts, "packet i/o", "<id>", "write manifest cert packets to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  manifest_id m_id;
  vector< manifest<cert> > certs;

  m_id = hexenc<id>(args[0]);
  app.db.get_manifest_certs(m_id, certs);
  for (size_t i = 0; i < certs.size(); ++i)
    pw.consume_manifest_cert(certs[i]);
  guard.commit();
}

CMD(fcerts, "packet i/o", "<id>", "write file cert packets to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);

  file_id f_id;
  vector< file<cert> > certs;

  f_id = hexenc<id>(args[0]);
  app.db.get_file_certs(f_id, certs);
  for (size_t i = 0; i < certs.size(); ++i)
    pw.consume_file_cert(certs[i]);
  guard.commit();
}

CMD(pubkey, "packet i/o", "<id>", "write public key packet to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);
  rsa_keypair_id ident(args[0]);
  base64< rsa_pub_key > key;
  app.db.get_key(ident, key);
  pw.consume_public_key(ident, key);
  guard.commit();
}


CMD(privkey, "packet i/o", "<id>", "write private key packet to stdout")
{
  if (args.size() != 1)
    throw usage();

  transaction_guard guard(app.db);
  packet_writer pw(cout);
  rsa_keypair_id ident(args[0]);
  base64< arc4<rsa_priv_key> > key;
  app.db.get_key(ident, key);
  pw.consume_private_key(ident, key);
  guard.commit();
}


CMD(read, "packet i/o", "", "read packets from stdin")
{
  transaction_guard guard(app.db);
  packet_db_writer dbw(app, true);
  size_t count = read_packets(cin, dbw);
  if (count == 0)
    throw oops("no packets found on stdin");
  else
    cerr << "read " << count << " packets" << endl;
  guard.commit();
}


CMD(agraph, "graph visualization", "", "dump ancestry graph to stdout")
{
  vector< manifest<cert> > certs;
  transaction_guard guard(app.db);
  app.db.get_manifest_certs(ancestor_cert_name, certs);
  set<string> nodes;
  vector< pair<string, string> > edges;
  for(vector< manifest<cert> >::iterator i = certs.begin();
      i != certs.end(); ++i)
    {
      gzip<cert_value> gz;
      cert_value tv;
      decode_base64(i->inner().value, gz);
      decode_gzip(gz, tv);
      nodes.insert(tv());
      nodes.insert(i->inner().ident());
      edges.push_back(make_pair(tv(), i->inner().ident()));
    }
  cout << "graph: " << endl << "{" << endl; // open graph
  for (set<string>::iterator i = nodes.begin(); i != nodes.end();
       ++i)
    {
      cout << "node: { title : \"" << *i << "\"}" << endl;
    }
  for (vector< pair<string,string> >::iterator i = edges.begin(); i != edges.end();
       ++i)
    {
      cout << "edge: { sourcename : \"" << i->first << "\"" << endl
	   << "        targetname : \"" << i->second << "\" }" << endl;
    }
  cout << "}" << endl << endl; // close graph
  guard.commit();
}

CMD(fetch, "nntp", "[server] [newsgroup]", "fetch recent changes from netnews")
{

  if (args.size() > 2)
    throw usage();

  map< pair<server,newsgroup>, unsigned long> seqs;

  vector<server> servs;
  app.db.get_read_servers(servs);
  for (vector<server>::const_iterator i = servs.begin();
       i != servs.end(); ++i)
    {
      if (args.size() > 0
	  && (*i)() != args[0])
	continue;
      vector<newsgroup> grps;
      app.db.get_groups(*i,grps);
      for (vector<newsgroup>::const_iterator j = grps.begin();
	   j != grps.end(); ++j)
	{
	  if (args.size() > 1
	      && (*j)() != args[1])
	    continue;
	  unsigned long seq;
	  app.db.get_seq(*i, *j, seq);
	  I(seqs.find(make_pair(*i,*j)) == seqs.end());
	  seqs.insert(make_pair(make_pair(*i,*j),seq));
	}
    }

  if (args.size() == 2)
    {
      pair<server,newsgroup> p = make_pair(server(args[0]),newsgroup(args[1]));
      if (seqs.find(p) == seqs.end())
	seqs.insert(make_pair(p,0));
    }

  // we now have a set of server/newsgroup/seq triples we want to update.

  unsigned long total = 0;
  packet_db_writer dbw(app);
  for (map< pair<server,newsgroup>, unsigned long>::const_iterator p = seqs.begin();
       p != seqs.end(); ++p)
    {
      string newsgroup = p->first.second();
      string host = p->first.first();
      int port = 119;
      size_t colon = host.find(':');
      unsigned long seq = p->second;
      unsigned long initial_seq = seq;
      if (colon != string::npos)
	{
	  port = boost::lexical_cast<unsigned long>(host.substr(colon+1));
	  host = host.substr(0, colon);
	}      
      transaction_guard guard(app.db);      
      L("connecting to %s:%d to fetch articles from %s starting with %lu\n",
	host.c_str(), port, newsgroup.c_str(), seq);
      fetch_articles(newsgroup, seq, dbw, host, port);
      if (initial_seq != seq)
	app.db.set_seq(p->first.first, p->first.second, seq);      
      guard.commit();
      if (seq > initial_seq)
	total += (seq - initial_seq);
    }
  cout << "advanced " << total << " article sequence numbers" << endl;
}

CMD(post, "nntp", "", "post queued changes to netnews")
{
  if (args.size() != 0)
    throw usage();

  stringstream ss;
  
  vector<server> servers;
  app.db.get_queued_servers(servers);

  if (servers.size() == 0)
    {
      cout << "no queued messages" << endl;
      return;
    }
  
  int num_posts = 0;

  for (vector<server>::const_iterator serv = servers.begin();
       serv != servers.end(); ++serv)
    {
      vector<posting> posts;
      app.db.get_queued_postings(*serv, posts);

      for (vector<posting>::const_iterator post = posts.begin();
	   post != posts.end(); ++post)
	{
	  transaction_guard guard(app.db);

	  string from, host(post->serv()), port("119"), group(post->group());
	  {
	    size_t colon = host.find(':');
	    if (colon != string::npos)
	      {
		port = host.substr(colon+1);
		host = host.substr(0, colon);
	      }
	  }

	  L("posting to %s : %s , group %s\n",
	    host.c_str(), port.c_str(), group.c_str());

	  if (!app.lua.hook_get_sender(post->serv, post->group, from))
	    {
	      L("no sender name found, skipping\n");
	    }
	  else
	    {
	      L("posting as '%s'\n", from.c_str());
	    }
	  
	  packet_writer pw(ss);	  
	  manifest_id m_old_id, m_new_id; 
	  manifest_data m_old_data, m_new_data;
	  manifest_map m_old, m_new;
	  patch_set ps;      
	  app.db.get_manifest_version(post->old_id, m_old_data);
	  app.db.get_manifest_version(post->new_id, m_new_data);
	  read_manifest_map(m_old_data, m_old);
	  read_manifest_map(m_new_data, m_new);
	  manifests_to_patch_set(m_old, m_new, app, ps);
	  patch_set_to_packets(ps, app, pw);  
	  guard.commit();

	  // body built
	  
	  string body = ss.str();
	  
	  post_article(group, from, ("[MT] " 
				     + post->old_id.inner()() + " -> " 
				     + post->new_id.inner()()), 
		       body, host, boost::lexical_cast<int>(port));
	  ++num_posts;
	}
    }
  cout << "posted " << num_posts << " articles" << endl;
	  
}


CMD(rcs_import, "rcs", "<rcsfile> ...", "import all versions in RCS files")
{
  transaction_guard guard(app.db);
  for (vector<string>::const_iterator i = args.begin();
       i != args.end(); ++i)
    {
      import_rcs_file(fs::path(*i), app.db);
    }
  guard.commit();
}


CMD(cvs_import, "rcs", "<cvsroot>", "import all versions in CVS repository")
{
  I(args.size() == 1);
  import_cvs_repo(fs::path(args.at(0)), app);
}


}; // namespace commands
