// copyright (C) 2002, 2003 graydon hoare <graydon@pobox.com>
// all rights reserved.
// licensed to the public under the terms of the GNU GPL (>= 2)
// see the file COPYING for details

#include <iterator>
#include <list>
#include <set>
#include <sstream>
#include <vector>

#include <stdarg.h>

#include <sqlite.h>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>

#include "cleanup.hh"
#include "database.hh"
#include "nntp_tasks.hh"
#include "sanity.hh"
#include "cert.hh"
#include "transforms.hh"
#include "vocab.hh"
#include "xdelta.hh"

// this file defines a public, typed interface to the database.
// the database class encapsulates all knowledge about sqlite,
// the schema, and all SQL statements used to access the schema.
//
// see file schema.sql for the text of the schema.

using boost::shared_ptr;
using boost::lexical_cast;

int const one_row = 1;
int const one_col = 1;
int const any_rows = -1;
int const any_cols = -1;
unsigned int const log_line_sz = 70;

extern "C" {
  // strangely this isn't declared, even though it's present in my sqlite.
  char *sqlite_vmprintf(const char *zFormat, va_list);
}

// defined in schema.sql, objcopied in 
extern char const _binary_schema_null_start[];

database::database(fs::path const & fn) :
  filename(fn),
  schema("1.0"),
  __sql(NULL),
  transaction_level(0)
{}

void database::check_schema()
{
  bool reinit_schema = false;

  try 
    {
      results res;
      fetch(res, one_col, one_row, "SELECT version FROM schema_version");
      I(res[0][0] == schema);
    }  
  catch (oops & dbe)  
    {
      reinit_schema = true;
    }
  
  if (reinit_schema)
    {
      results res;
      fetch(res, one_col, any_rows, "SELECT name FROM sqlite_master WHERE type = 'table'");
      for (unsigned int i = 0; i < res.size(); ++i)
	execute("DROP TABLE '%q'", res[i][0].c_str());
      execute(_binary_schema_null_start);
      execute("INSERT INTO schema_version VALUES('%q')", schema.c_str());
    }
}

struct sqlite * database::sql()
{
  if (! __sql)
    {
      char * errmsg;
      __sql = sqlite_open(filename.string().c_str(), 0755, &errmsg);
      if (! __sql)
	throw oops(string("could not open database: ") + filename.string());
      check_schema();
    }
  return __sql;
}

void database::ensure_open()
{
  sqlite *s = sql();
  I(s != NULL);
}

database::~database() 
{
  if (__sql)
    {
      sqlite_close(__sql);
      __sql = 0;
    }
}

static void assert_sqlite_ok(int res)
{
  switch (res)
    {      
    case SQLITE_OK: 
      break;

    case SQLITE_ERROR:
      throw oops("SQL error or missing database");
      break;

    case SQLITE_INTERNAL:
      throw oops("An internal logic error in SQLite");
      break;

    case SQLITE_PERM:
      throw oops("Access permission denied");
      break;

    case SQLITE_ABORT:
      throw oops("Callback routine requested an abort");
      break;

    case SQLITE_BUSY:
      throw oops("The database file is locked");
      break;

    case SQLITE_LOCKED:
      throw oops("A table in the database is locked");
      break;

    case SQLITE_NOMEM:
      throw oops("A malloc() failed");
      break;

    case SQLITE_READONLY:
      throw oops("Attempt to write a readonly database");
      break;

    case SQLITE_INTERRUPT:
      throw oops("Operation terminated by sqlite_interrupt()");
      break;

    case SQLITE_IOERR:
      throw oops("Some kind of disk I/O error occurred");
      break;

    case SQLITE_CORRUPT:
      throw oops("The database disk image is malformed");
      break;

    case SQLITE_NOTFOUND:
      throw oops("(Internal Only) Table or record not found");
      break;

    case SQLITE_FULL:
      throw oops("Insertion failed because database is full");
      break;

    case SQLITE_CANTOPEN:
      throw oops("Unable to open the database file");
      break;

    case SQLITE_PROTOCOL:
      throw oops("database lock protocol error");
      break;

    case SQLITE_EMPTY:
      throw oops("(Internal Only) database table is empty");
      break;

    case SQLITE_SCHEMA:
      throw oops("The database schema changed");
      break;

    case SQLITE_TOOBIG:
      throw oops("Too much data for one row of a table");
      break;

    case SQLITE_CONSTRAINT:
      throw oops("Abort due to contraint violation");
      break;

    case SQLITE_MISMATCH:
      throw oops("Data type mismatch");
      break;

    case SQLITE_MISUSE:
      throw oops("Library used incorrectly");
      break;

    default:
      throw oops(string("Unknown DB result code: ") + lexical_cast<string>(res));
      break;
    }
}

void database::execute(char const * query, ...)
{
  va_list ap;
  int res;
  char * errmsg = NULL;

  va_start(ap, query);

  // log it
  char * formatted = sqlite_vmprintf(query, ap);
  string qq(formatted);
  if (qq.size() > log_line_sz) 
    qq = qq.substr(0, log_line_sz) + string(" ...");
  L("db.execute(\"%s\")\n", qq.c_str());
  sqlite_freemem(formatted);

  // do it
  res = sqlite_exec_vprintf(sql(), query, NULL, NULL, &errmsg, ap);

  va_end(ap);

  if (errmsg)
    throw oops(string("sqlite exec error ") + errmsg);

  assert_sqlite_ok(res);

}

void database::fetch(results & res, 
	   int const want_cols, 
	   int const want_rows, 
	   char const * query, ...)
{
  char ** result = NULL;
  int nrow;
  int ncol;
  char * errmsg = NULL;
  int rescode;

  va_list ap;
  res.clear();
  res.resize(0);
  va_start(ap, query);

  // log it
  char * formatted = sqlite_vmprintf(query, ap);
  string qq(formatted);
  if (qq.size() > log_line_sz) 
    qq = qq.substr(0, log_line_sz) + string(" ...");
  L("db.fetch(\"%s\")\n", qq.c_str());
  sqlite_freemem(formatted);

  // do it
  rescode = sqlite_get_table_vprintf(sql(), query, &result, &nrow, &ncol, &errmsg, ap);

  va_end(ap);

  cleanup_ptr<char **, void> 
    result_guard(result, &sqlite_free_table);

  string ctx = string("db query [") + string(query) + "]: ";

  if (errmsg)
    throw oops(ctx + string("sqlite error ") + errmsg);
  assert_sqlite_ok(rescode);

  if (want_cols == 0 && ncol == 0) return;
  if (want_rows == 0 && nrow == 0) return;
  if (want_cols == any_rows && ncol == 0) return;
  if (want_rows == any_rows && nrow == 0) return;

  if (want_cols != any_cols &&
      ncol != want_cols)
    throw oops(ctx + "wanted " + lexical_cast<string>(want_cols) 
	       + " columns, got " +  lexical_cast<string>(ncol));

  if (want_rows != any_rows &&
      nrow != want_rows)
    throw oops(ctx + "wanted " + lexical_cast<string>(want_rows)
	       + " rows, got " +  lexical_cast<string>(nrow));

  if (!result)
    throw oops(ctx + "null result set");

  for (int i = 0; i < ncol; ++i) 
    if (!result[i])
      throw oops(ctx + "null column name");

  for (int row = 0; row < nrow; ++row) 
    {
      vector<string> rowvec;
      for (int col = 0; col < ncol; ++col)
	{
	  int i = ((1 + row) * ncol) + col;
	  if (!result[i])
	    throw oops(ctx + "null result value");
	  else
	    rowvec.push_back(result[i]);
	}
      res.push_back(rowvec);
    }
}

// general application-level logic

void database::set_filename(fs::path const & file)
{
  if (__sql)
    {
      throw oops("cannot change filename to " + file.string() + " while db is open");
    }
  filename = file;
}

void database::begin_transaction() 
{
  if (transaction_level == 0)
      execute("BEGIN");
  transaction_level++;
}

void database::commit_transaction()
{
  if (transaction_level == 1)
    execute("COMMIT");
  transaction_level--;
}

void database::rollback_transaction()
{
  if (transaction_level == 1)
    execute("ROLLBACK");
  transaction_level--;
}


bool database::exists(hexenc<id> const & ident,
		      string const & table)
{
  results res;
  fetch(res, one_col, any_rows, "SELECT id FROM '%q' WHERE id = '%q'",
	table.c_str(), ident().c_str());
  I((res.size() == 1) || (res.size() == 0));
  return res.size() == 1;
}


bool database::delta_exists(hexenc<id> const & ident,
			    string const & table)
{
  results res;
  fetch(res, one_col, any_rows, "SELECT id FROM '%q' WHERE id = '%q'",
	table.c_str(), ident().c_str());
  return res.size() > 0;
}

bool database::delta_exists(hexenc<id> const & ident,
			    hexenc<id> const & base,
			    string const & table)
{
  results res;
  fetch(res, one_col, any_rows, "SELECT id FROM '%q' WHERE id = '%q' AND base = '%q'",
	table.c_str(), ident().c_str(), base().c_str());
  I((res.size() == 1) || (res.size() == 0));
  return res.size() == 1;
}

int database::count(string const & table)
{
  results res;
  fetch(res, one_col, one_row, "SELECT COUNT(*) FROM '%q'", table.c_str());
  return lexical_cast<int>(res[0][0]);  
}

void database::get(hexenc<id> const & ident,
		   base64< gzip<data> > & dat,
		   string const & table)
{
  results res;
  fetch(res, one_col, one_row,
	"SELECT data FROM '%q' WHERE id = '%q'", 
	table.c_str(), ident().c_str());

  // consistency check
  base64<gzip<data> > rdata(res[0][0]);
  hexenc<id> tid;
  calculate_ident(rdata, tid);
  I(tid == ident);

  dat = rdata;
}

void database::get_delta(hexenc<id> const & ident,
			 hexenc<id> const & base,
			 base64< gzip<delta> > & del,
			 string const & table)
{
  I(ident() != "");
  I(base() != "");
  results res;
  fetch(res, one_col, one_row,
	"SELECT delta FROM '%q' WHERE id = '%q' AND base = '%q'", 
	table.c_str(), ident().c_str(), base().c_str());
  del = res[0][0];
}

void database::put(hexenc<id> const & ident,
		   base64< gzip<data> > const & dat,
		   string const & table)
{
  // consistency check
  I(ident() != "");
  hexenc<id> tid;
  calculate_ident(dat, tid);
  I(tid == ident);
  
  execute("INSERT INTO '%q' VALUES('%q', '%q')", 
	  table.c_str(), ident().c_str(), dat().c_str());
}


void database::put_delta(hexenc<id> const & ident,
			 hexenc<id> const & base,
			 base64<gzip<delta> > const & del,
			 string const & table)
{
  // nb: delta schema is (id, base, delta)
  I(ident() != "");
  I(base() != "");
  execute("INSERT INTO '%q' VALUES('%q', '%q', '%q')", 
	  table.c_str(), 
	  ident().c_str(), base().c_str(), del().c_str());
}

void database::get_version(hexenc<id> const & ident,
			   base64< gzip<data> > & dat,
			   string const & data_table,
			   string const & delta_table)
{
  I(ident() != "");
  if (exists(ident, data_table))
    {
      // easy path
      get(ident, dat, data_table);
    }
  else
    {
      // tricky path

      // we start from the file we want to reconstruct and work *forwards*
      // through the database, until we get to a full data object. we then
      // trace back through the list of edges we followed to get to the data
      // object, applying reverse deltas.
      //
      // the effect of this algorithm is breadth-first search, backwards
      // through the storage graph, to discover a forwards shortest path, and
      // then following that shortest path with delta application.
      //
      // we used to do this with the boost graph library, but it invovled
      // loading too much of the storage graph into memory at any moment. this
      // imperative version only loads the descendents of the reconstruction
      // node, so it much cheaper in terms of memory.
      //
      // we also maintain a cycle-detecting set, just to be safe
      
      L("reconstructing %s in %s\n", ident().c_str(), delta_table.c_str());
      I(delta_exists(ident, delta_table));
      
      // nb: an edge map goes in the direction of the
      // delta, *not* the direction we discover things in,
      // i.e. each map is of the form [newid] -> [oldid]

      typedef map< hexenc<id>, hexenc<id> > edgemap;
      list< shared_ptr<edgemap> > paths;

      set< hexenc<id> > frontier, cycles;
      frontier.insert(ident);

      bool found_root = false;
      hexenc<id> root("");

      while (! found_root)
	{
	  set< hexenc<id> > next_frontier;
	  shared_ptr<edgemap> frontier_map(new edgemap());

	  for (set< hexenc<id> >::const_iterator i = frontier.begin();
	       i != frontier.end(); ++i)
	    {
	      if (exists(*i, data_table))
		{
		  root = *i;
		  found_root = true;
		  break;
		}
	      else
		{
		  cycles.insert(*i);
		  results res;
		  fetch(res, one_col, any_rows, "SELECT base from '%q' WHERE id = '%q'",
			delta_table.c_str(), (*i)().c_str());
		  for (size_t k = 0; k < res.size(); ++k)
		    {
		      hexenc<id> const nxt(res[k][0]);

		      if (cycles.find(nxt) != cycles.end())
			throw oops("cycle in table '" + delta_table + "', at node " 
				   + (*i)() + " <- " + nxt());

		      next_frontier.insert(nxt);

		      if (frontier_map->find(nxt) == frontier_map->end())
			{
			  L("inserting edge: %s <- %s\n", (*i)().c_str(), nxt().c_str());
			  frontier_map->insert(make_pair(nxt, *i));
			}
		      else
			L("skipping merge edge %s <- %s\n", (*i)().c_str(), nxt().c_str());
		    }
		}
	    }
	  if (!found_root)
	    {
	      frontier = next_frontier;
	      paths.push_front(frontier_map);
	    }
	}

      // path built, now all we need to do is follow it back

      I(found_root);
      I(root() != "");
      base64< gzip<data> > begin_packed;
      data begin;      
      get(root, begin_packed, data_table);
      unpack(begin_packed, begin);
      hexenc<id> curr = root;

      boost::shared_ptr<delta_applicator> app = new_piecewise_applicator();
      app->begin(begin());
      
      for (list< shared_ptr<edgemap> >::const_iterator p = paths.begin();
	   p != paths.end(); ++p)
	{
	  shared_ptr<edgemap> i = *p;
	  I(i->find(curr) != i->end());
	  hexenc<id> const nxt = i->find(curr)->second;

	  L("following delta %s -> %s\n", curr().c_str(), nxt().c_str());
	  base64< gzip<delta> > del_packed;
	  get_delta(nxt, curr, del_packed, delta_table);
	  delta del;
	  unpack(del_packed, del);
	  apply_delta (app, del());
	  app->next();
	  curr = nxt;
	}

      string tmp;
      app->finish(tmp);
      data end(tmp);

      hexenc<id> final;
      calculate_ident(end, final);
      I(final == ident);
      pack(end, dat);
    }
}


void database::drop(hexenc<id> const & ident, 
		    string const & table)
{
  execute("DELETE FROM '%q' WHERE id = '%q'",  
	  table.c_str(),
	  ident().c_str());
}


void database::put_version(hexenc<id> const & old_id,
			   hexenc<id> const & new_id,
			   base64< gzip<delta> > const & del,
			   string const & data_table,
			   string const & delta_table)
{

  base64< gzip<data> > old_data, new_data;
  base64< gzip<delta> > reverse_delta;
  
  get_version(old_id, old_data, data_table, delta_table);
  patch(old_data, del, new_data);
  diff(new_data, old_data, reverse_delta);
      
  transaction_guard guard(*this);
  if (exists(old_id, data_table))
    {
      // descendent of a head version replaces the head, therefore old head
      // must be disposed of
      drop(old_id, data_table);
    }
  put(new_id, new_data, data_table);
  put_delta(old_id, new_id, reverse_delta, delta_table);
  guard.commit();
}



// ------------------------------------------------------------
// --                                                        --
// --              public interface follows                  --
// --                                                        --
// ------------------------------------------------------------

bool database::file_version_exists(file_id const & id)
{
  return delta_exists(id.inner(), "file_deltas") 
    || exists(id.inner(), "files");
}

bool database::manifest_version_exists(manifest_id const & id)
{
  return delta_exists(id.inner(), "manifest_deltas") 
    || exists(id.inner(), "manifests");
}

void database::get_file_version(file_id const & id,
				file_data & dat)
{
  base64< gzip<data> > tmp;
  get_version(id.inner(), tmp, "files", "file_deltas");
  dat = tmp;
}

void database::get_manifest_version(manifest_id const & id,
				    manifest_data & dat)
{
  base64< gzip<data> > tmp;
  get_version(id.inner(), tmp, "manifests", "manifest_deltas");
  dat = tmp;
}

void database::put_file(file_id const & id,
			file_data const & dat)
{
  put(id.inner(), dat.inner(), "files");
}

void database::put_file_version(file_id const & old_id,
				file_id const & new_id,
				file_delta const & del)
{
  put_version(old_id.inner(), new_id.inner(), del.inner(), 
	      "files", "file_deltas");
}


void database::put_manifest(manifest_id const & id,
			    manifest_data const & dat)
{
  put(id.inner(), dat.inner(), "manifests");
}

void database::put_manifest_version(manifest_id const & old_id,
				    manifest_id const & new_id,
				    manifest_delta const & del)
{
  put_version(old_id.inner(), new_id.inner(), del.inner(), 
	      "manifests", "manifest_deltas");
}

// crypto key management

void database::get_key_ids(string const & pattern,
			   vector<rsa_keypair_id> & pubkeys,
			   vector<rsa_keypair_id> & privkeys)
{
  pubkeys.clear();
  privkeys.clear();
  results res;

  if (pattern != "")
    fetch(res, one_col, any_rows, 
	  "SELECT id from public_keys WHERE id GLOB '%q'",
	  pattern.c_str());
  else
    fetch(res, one_col, any_rows, 
	  "SELECT id from public_keys");

  for (size_t i = 0; i < res.size(); ++i)
    pubkeys.push_back(res[i][0]);

  if (pattern != "")
    fetch(res, one_col, any_rows, 
	  "SELECT id from private_keys WHERE id GLOB '%q'",
	  pattern.c_str());
  else
    fetch(res, one_col, any_rows, 
	  "SELECT id from private_keys");

  for (size_t i = 0; i < res.size(); ++i)
    privkeys.push_back(res[i][0]);
}

void database::get_private_keys(vector<rsa_keypair_id> & privkeys)
{
  privkeys.clear();
  results res;
  fetch(res, one_col, any_rows,  "SELECT id from private_keys");
  for (size_t i = 0; i < res.size(); ++i)
    privkeys.push_back(res[i][0]);
}


bool database::public_key_exists(rsa_keypair_id const & id)
{
  results res;
  fetch(res, one_col, any_rows, 
	"SELECT id FROM public_keys WHERE id = '%q'",
	id().c_str());
  I((res.size() == 1) || (res.size() == 0));
  if (res.size() == 1) 
    return true;
  return false;
}

bool database::private_key_exists(rsa_keypair_id const & id)
{
  results res;
  fetch(res, one_col, any_rows,
	"SELECT id FROM private_keys WHERE id = '%q'",
	id().c_str());
  I((res.size() == 1) || (res.size() == 0));
  if (res.size() == 1)
    return true;
  return false;
}

bool database::key_exists(rsa_keypair_id const & id)
{
  return public_key_exists(id) || private_key_exists(id);
}


void database::get_key(rsa_keypair_id const & pub_id, 
		       base64<rsa_pub_key> & pub_encoded)
{
  results res;
  fetch(res, one_col, one_col, 
	"SELECT keydata FROM public_keys where id = '%q'", 
	pub_id().c_str());
  pub_encoded = res[0][0];
}

void database::get_key(rsa_keypair_id const & priv_id, 
		       base64< arc4<rsa_priv_key> > & priv_encoded)
{
  results res;
  fetch(res, one_col, one_col, 
	"SELECT keydata FROM private_keys where id = '%q'", 
	priv_id().c_str());
  priv_encoded = res[0][0];
}


void database::put_key(rsa_keypair_id const & pub_id, 
		       base64<rsa_pub_key> const & pub_encoded)
{
  execute("INSERT INTO public_keys VALUES('%q', '%q')", 
	  pub_id().c_str(), pub_encoded().c_str());
}

void database::put_key(rsa_keypair_id const & priv_id, 
		       base64< arc4<rsa_priv_key> > const & priv_encoded)
{
  execute("INSERT INTO private_keys VALUES('%q', '%q')", 
	  priv_id().c_str(), priv_encoded().c_str());
}

void database::put_key_pair(rsa_keypair_id const & id, 
			    base64<rsa_pub_key> const & pub_encoded,
			    base64< arc4<rsa_priv_key> > const & priv_encoded)
{
  transaction_guard guard(*this);
  put_key(id, pub_encoded);
  put_key(id, priv_encoded);
  guard.commit();
}


// cert management

bool database::cert_exists(cert const & t,
			  string const & table)
{
  results res;
  fetch(res, 1, any_rows,
	"SELECT id FROM '%q' WHERE id = '%q'"
	" AND name = '%q' AND value = '%q'" 
	" AND keypair = '%q' AND signature = '%q'",
	table.c_str(),
	t.ident().c_str(),
	t.name().c_str(),
	t.value().c_str(),
	t.key().c_str(),
	t.sig().c_str());
  I(res.size() == 0 || res.size() == 1);
  return res.size() == 1;
}

void database::put_cert(cert const & t,
		       string const & table)
{
  execute("INSERT INTO '%q' VALUES('%q', '%q', '%q', '%q', '%q')", 
	  table.c_str(),
	  t.ident().c_str(),
	  t.name().c_str(), 
	  t.value().c_str(),
	  t.key().c_str(),
	  t.sig().c_str());
}

void database::results_to_certs(results const & res,
			       vector<cert> & certs)
{
  certs.clear();
  for (size_t i = 0; i < res.size(); ++i)
    {
      cert t;
      t = cert(hexenc<id>(res[i][0]), 
	      cert_name(res[i][1]),
	      base64< gzip<cert_value> >(res[i][2]),
	      rsa_keypair_id(res[i][3]),
	      hexenc<rsa_sha1_signature>(res[i][4]));
      certs.push_back(t);
    }
}

void database::get_certs(hexenc<id> const & ident, 
			vector<cert> & certs,			
			string const & table)
{
  results res;
  fetch(res, 5, any_rows, 
	"SELECT id, name, value, keypair, signature FROM '%q' WHERE id = '%q'", 
	table.c_str(), 	
	ident().c_str());
  results_to_certs(res, certs);
}


void database::get_certs(cert_name const & name, 	      
			vector<cert> & certs,
			string const & table)
{
  results res;
  fetch(res, 5, any_rows, 
	"SELECT id, name, value, keypair, signature FROM '%q' WHERE name = '%q'", 
	table.c_str(), 	
	name().c_str());
  results_to_certs(res, certs);
}


void database::get_certs(hexenc<id> const & ident, 
			cert_name const & name, 	      
			vector<cert> & certs,
			string const & table)
{
  results res;
  fetch(res, 5, any_rows, 
	"SELECT id, name, value, keypair, signature FROM '%q' WHERE id = '%q' AND name = '%q'", 
	table.c_str(), 	
	ident().c_str(),
	name().c_str());
  results_to_certs(res, certs);
}

void database::get_certs(cert_name const & name,
			base64< gzip<cert_value> > const & val, 
			vector<cert> & certs,
			string const & table)
{
  results res;
  fetch(res, 5, any_rows, 
	"SELECT id, name, value, keypair, signature FROM '%q' WHERE name = '%q' AND value = '%q'", 
	table.c_str(), 	
	name().c_str(),
	val().c_str());
  results_to_certs(res, certs);
}


void database::get_certs(hexenc<id> const & ident, 
			cert_name const & name, 	      
			base64< gzip<cert_value> > const & value,
			vector<cert> & certs,
			string const & table)
{
  results res;
  fetch(res, 5, any_rows, 
	"SELECT id, name, value, keypair, signature FROM '%q' WHERE id = '%q' AND name = '%q' AND value = '%q'", 
	table.c_str(), 	
	ident().c_str(),
	name().c_str(),
	value().c_str());
  results_to_certs(res, certs);
}




bool database::manifest_cert_exists(manifest<cert> const & cert)
{ return cert_exists(cert.inner(), "manifest_certs"); }

bool database::file_cert_exists(file<cert> const & cert)
{ return cert_exists(cert.inner(), "file_certs"); }

void database::put_manifest_cert(manifest<cert> const & cert)
{ put_cert(cert.inner(), "manifest_certs"); }

void database::put_file_cert(file<cert> const & cert)
{ put_cert(cert.inner(), "file_certs"); }

void database::get_file_certs(cert_name const & name, 
			     vector< file<cert> > & ts)
{
  vector<cert> certs;
  get_certs(name, certs, "file_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_file_certs(file_id const & id, 
			     cert_name const & name, 
			     vector< file<cert> > & ts)
{
  vector<cert> certs;
  get_certs(id.inner(), name, certs, "file_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));    
}

void database::get_file_certs(cert_name const & name,
			     base64< gzip<cert_value> > const & val, 
			     vector< file<cert> > & ts)
{
  vector<cert> certs;
  get_certs(name, val, certs, "file_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_file_certs(file_id const & id, 
			     cert_name const & name,
			     base64< gzip<cert_value> > const & val, 
			     vector< file<cert> > & ts)
{
  vector<cert> certs;
  get_certs(id.inner(), name, val, certs, "file_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_file_certs(file_id const & id, 
			     vector< file<cert> > & ts)
{ 
  vector<cert> certs;
  get_certs(id.inner(), certs, "file_certs"); 
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));
}






void database::get_manifest_certs(cert_name const & name, 
				 vector< manifest<cert> > & ts)
{
  vector<cert> certs;
  get_certs(name, certs, "manifest_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_manifest_certs(manifest_id const & id, 
				 cert_name const & name, 
				 vector< manifest<cert> > & ts)
{
  vector<cert> certs;
  get_certs(id.inner(), name, certs, "manifest_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_manifest_certs(manifest_id const & id, 
				 cert_name const & name,
				 base64< gzip<cert_value> > const & val, 
				 vector< manifest<cert> > & ts)
{
  vector<cert> certs;
  get_certs(id.inner(), name, val, certs, "manifest_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_manifest_certs(cert_name const & name,
				 base64< gzip<cert_value> > const & val, 
				 vector< manifest<cert> > & ts)
{
  vector<cert> certs;
  get_certs(name, val, certs, "manifest_certs");
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));  
}

void database::get_manifest_certs(manifest_id const & id, 
				 vector< manifest<cert> > & ts)
{ 
  vector<cert> certs;
  get_certs(id.inner(), certs, "manifest_certs"); 
  ts.clear();
  copy(certs.begin(), certs.end(), back_inserter(ts));
}


// netnews stuff

void database::queue_to_send(posting const & post)
{
  I(manifest_version_exists(post.old_id));
  I(manifest_version_exists(post.new_id));
  execute("INSERT INTO posting_queue VALUES ('%q','%q','%q','%q')",
	  post.serv().c_str(),
	  post.group().c_str(),
	  post.old_id.inner()().c_str(),
	  post.new_id.inner()().c_str());
}

void database::get_queued_servers(vector<server> & servers)
{
  results res;
  fetch(res, one_col, any_rows, "SELECT server FROM posting_queue GROUP BY server");
  servers.clear();
  for (size_t i = 0; i < res.size(); ++i)
    servers.push_back(server(res[i][0]));
}

void database::get_queued_postings(server const & serv, 
				   vector<posting> & posts)
{
  results res;
  fetch(res, 4, any_rows, 
	"SELECT server, newsgroup, old_id, new_id FROM posting_queue WHERE server = '%q'",
	serv().c_str());
  posts.clear();
  for (size_t i = 0; i < res.size(); ++i)
    {
      posting post = posting(server(res[i][0]),
			     newsgroup(res[i][1]),
			     manifest_id(res[i][2]),
			     manifest_id(res[i][3]));
      I(manifest_version_exists(post.old_id));
      I(manifest_version_exists(post.new_id));      
      posts.push_back(post);
    }
}

void database::delete_posting(posting const & post)
{
  execute("DELETE FROM posting_queue WHERE server = '%q' AND newsgroup = '%q' AND old_id = '%q' AND new_id = '%q'",
	  post.serv().c_str(),
	  post.group().c_str(),
	  post.old_id.inner()().c_str(),
	  post.new_id.inner()().c_str());
}


void database::get_read_servers(vector<server> & servers)
{
  results res;
  fetch(res, one_col, any_rows,
	"SELECT server FROM sequence_numbers GROUP BY server");
  servers.clear();
  for (size_t i = 0; i < res.size(); ++i)
    servers.push_back(server(res[i][0]));
}

void database::get_groups(server const & serv,
		vector<newsgroup> & groups)
{
  results res;
  fetch(res, one_col, any_rows,
	"SELECT newsgroup FROM sequence_numbers WHERE server = '%q'",
	serv().c_str());
  groups.clear();
  for (size_t i = 0; i < res.size(); ++i)
    groups.push_back(newsgroup(res[i][0]));
}

void database::get_seq(server const & serv,
		       newsgroup const & group,
		       unsigned long & seq)
{
  results res;
  fetch(res, any_cols, any_rows,
	"SELECT seq FROM sequence_numbers WHERE server = '%q' AND newsgroup = '%q'",
	serv().c_str(), group().c_str());
  I(res.size() == 1 || res.size() == 0);  
  if (res.size() == 0)
    {
      execute("INSERT INTO sequence_numbers VALUES ('%q','%q','0')",
	      serv().c_str(), group().c_str());
      seq = 0;
    }
  else
    {
      I(res[0].size() == 1);
      seq = lexical_cast<unsigned long>(res[0][0]);
    }
}

void database::set_seq(server const & serv,
		       newsgroup const & group,
		       unsigned long const & seq)
{
  unsigned long dummy_seq;
  get_seq(serv, group, dummy_seq);
  execute("UPDATE sequence_numbers SET seq = '%q' WHERE server = '%q' AND newsgroup = '%q'",
	  lexical_cast<string>(seq).c_str(),
	  serv().c_str(), group().c_str());  
}


// transaction guards

transaction_guard::transaction_guard(database & d) : committed(false), db(d) 
{
  db.begin_transaction();
}
transaction_guard::~transaction_guard()
{
  if (committed)
    db.commit_transaction();
  else
    db.rollback_transaction();
}

void transaction_guard::commit()
{
  committed = true;
}
