// copyright (C) 2002, 2003 graydon hoare <graydon@pobox.com>
// all rights reserved.
// licensed to the public under the terms of the GNU GPL (>= 2)
// see the file COPYING for details

extern "C" {
#include <lua.h>
#include <lualib.h>
}

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

#include "lua.hh"
#include "sanity.hh"
#include "vocab.hh"
#include "file_io.hh"
#include "app_state.hh"

// defined in std_hooks.lua, objcopied in 
extern char const _binary_std_hooks_null_start[];

#ifdef BUILD_UNIT_TESTS
extern char const _binary_test_hooks_null_start[];
#endif

using std::string;
using boost::lexical_cast;

lua_hooks::lua_hooks()
{
  st = lua_open (0);  
  I(st);
  lua_baselibopen(st);
  lua_iolibopen(st);
  lua_strlibopen(st);
  lua_mathlibopen(st);
}

lua_hooks::~lua_hooks()
{
  if (st)
    lua_close (st);
}

#ifdef BUILD_UNIT_TESTS
void lua_hooks::add_test_hooks()
{
  I(st);
  if (lua_dostring(st, _binary_test_hooks_null_start) != 0)
    throw oops("lua error while setting up testing hooks");
}
#endif

void lua_hooks::add_std_hooks()
{
  I(st);
  if (lua_dostring(st, _binary_std_hooks_null_start) != 0)
    throw oops("lua error while setting up standard hooks");
}

void lua_hooks::default_rcfilename(fs::path & file)
{
  char * home = getenv("HOME");
  if (!home)
    throw oops("no HOME directory in $ENV");
  file = fs::path(home) / fs::path(".monotonerc");
}

void lua_hooks::add_rcfile(fs::path const & rc)
{
  I(st);
  if (fs::exists(rc))
    if (lua_dofile(st, rc.string().c_str()) != 0)
      throw oops("lua error while loading " + rc.string());
}

template <typename RES>
static inline RES extract(struct lua_State * st);

template<>
static inline double extract<double>(struct lua_State * st)
{ return lua_tonumber(st, -1); }

template<>
static inline string extract<string>(struct lua_State * st)
{ return string(lua_tostring(st, -1), lua_strlen(st, -1)); }

template <typename T> static inline T dummy();
template <> static inline string dummy<string>() { return string(); }
template <> static inline double dummy<double>() { return 0.; }

template <typename RES>
pair<bool,RES> lua_hooks::call_hook(char const * name, int nargs, ...)
{
  va_list ap;

  I(st);
  va_start(ap, nargs);
  lua_getglobal(st, name);
  if (! lua_isfunction(st, -1))
    {
      lua_pop(st, 1);
      RES dum = dummy<RES>();
      L("lua hook: %s not found\n", name);
      return make_pair(false, dum);
    }

  int c = nargs;
  while (c--)
    {
      char * arg = va_arg(ap, char *);
      int len = va_arg(ap, int);
      lua_pushlstring(st, arg, len);
    }

  va_end(ap);

  L("calling lua hook %s\n", name);
  if (lua_call(st, nargs, 1) != 0)
    throw oops("lua error while running '" + string(name) + "' hook");

  RES res = extract<RES>(st);
  lua_pop(st, 1);

  return make_pair(true, res);
}

// concrete hooks

// nb: if you're hooking lua to return your passphrase, you don't care if we
// keep a couple extra temporaries of your passphrase around.
bool lua_hooks::hook_get_passphrase(rsa_keypair_id const & k, string & phrase)
{
  pair<bool,string> res = call_hook<string>("get_passphrase", 1, 
					    k().data(), k().size());
  phrase = res.second;
  return res.first;
}

bool lua_hooks::hook_persist_phrase_ok()
{
  pair<bool,double> res = call_hook<double>("persist_phrase_ok", 0); 
  return res.first && (res.second != 0.);
}


bool lua_hooks::hook_get_branch_key(cert_value const & branchname, rsa_keypair_id & k)
{
  pair<bool,string> res = call_hook<string>("get_branch_key", 1, 
					    branchname().data(), branchname().size());
  k = res.second;
  return res.first;
}

bool lua_hooks::hook_get_author(cert_value const & branchname, string & author)
{
  pair<bool,string> res = call_hook<string>("get_author", 1,
					    branchname().data(), branchname().size());
  author = res.second;
  return res.first;    
}

bool lua_hooks::hook_edit_comment(string const & commentary, string & result)
{
  pair<bool,string> res = call_hook<string>("edit_comment", 1,
					    commentary.data(), commentary.size());
  result = res.second;
  return res.first; 
}


bool lua_hooks::hook_get_sorter(string const & certname, string & sort_type)
{
  pair<bool,string> res = call_hook<string>("get_sorter", 1,
					    certname.data(), certname.size());
  sort_type = res.second;
  return res.first;    
}

bool lua_hooks::hook_ignore_file(file_path const & p)
{
  pair<bool,double> res = call_hook<double>("ignore_file", 1, 
					    p().data(), p().size());
  return res.first && (res.second != 0.);
}


bool lua_hooks::hook_non_blocking_rng_ok()
{
  pair<bool,double> res = call_hook<double>("non_blocking_rng_ok", 0);
  return res.first && (res.second != 0.);
}


bool lua_hooks::hook_get_newsserver(cert_value const & branchname, server & serv)
{
  pair<bool,string> res = call_hook<string>("get_newsserver", 1, 
					    branchname().data(), branchname().size());
  serv = res.second;
  return res.first;
}

bool lua_hooks::hook_get_newsgroup(cert_value const & branchname, newsgroup & grp)
{
  pair<bool,string> res = call_hook<string>("get_newsgroup", 1, 
					    branchname().data(), branchname().size());
  grp = res.second;
  return res.first;
}

bool lua_hooks::hook_get_sender(server const & serv, newsgroup const & grp, 
				string & sender)
{
  pair<bool,string> res = call_hook<string>("get_sender", 2, 
					    serv().data(), serv().size(),
					    grp().data(), grp().size());
  sender = res.second;
  return res.first;
}

bool lua_hooks::hook_merge2(data const & left, 
			    data const & right, 
			    data & result)
{
  pair<bool,string> res = call_hook<string>("merge2", 2, 
					    left().data(), left().size(),
					    right().data(), right().size());
  result = res.second;
  return res.first;
}

bool lua_hooks::hook_merge3(data const & ancestor, 
			    data const & left, 
			    data const & right, 
			    data & result)
{
  pair<bool,string> res = call_hook<string>("merge3", 3,
					    ancestor().data(), ancestor().size(),
					    left().data(), left().size(),
					    right().data(), right().size());
  result = res.second;
  return res.first;
}
